#
# Various utility procedures
#
#########################################################
# A procedure giving the version info of ERS
proc ERS_Version { } {
    return "ERS v 0.2.1 (11/03/99)"
}

# Some procedures for complaining about installation errors! Note that
# they are mainly meant for copy/paste-ing since they are invokes
# precisely when they cannot be found here...
proc ERS_Install_Error { } {

    tk_dialog .error "Installation Error" \
	"Don't know where is ERS.\nPlease set up the ERSHOME environment variable." {error} 0 {Ok}
}

proc ERS_Directory_Error { dirname } {

    tk_dialog .error "Installation Error" \
	"Can't find directory $dirname\nPlease check the ERSHOME environment variable." {error} 0 {Ok}

}

######################################################################
# ReportError -
# ReportWarning -
# ReportInfo -
# Procedures to pop up an information message. Could be more sophisticated
# and format their messages better.
# 
# Parameters:
# - title	the title to be given to the window
# - message	the message to be written inside.
#
proc ReportError { title message } {
    tk_dialog .error "$title" "$message" {error} 0 {Ok}
}

proc ReportWarning { title message } {
    tk_dialog .warning "$title" "$message" {warning} 0 {Ok}
}

proc ReportInfo { title message } {

    if { [info tclversion] > 7.5 } {
	tk_messageBox -type ok -title $title -icon info -message $message
    } else {
	tk_dialog .information "$title" "$message" {info} 0 {Dismiss}
    }
}

######################################################################
# Iota -
# A function for creating the list of consecutive numbers.
#
# Parameters:
# - n		the first index of the list
# - m		the last index of the list
#
proc iota { n m } {
    
    set res {}

    for { set i $n } { $i <= $m } { incr i 1 } {
	set res [concat $res $i]
    }

    return $res
}

######################################################################
# UpperCase -
# A function changing the first letter of a string to upper case
#
# Parameters:
# - s		the string to modify
#
proc UpperCase { s } {
    return "[string toupper [string range $s 0 0]][string range $s 1 end]"
}

######################################################################
# GetOption -
# A function for extracting the value of a named option in a list
# of option/value.
# Returns an empty string in no option is found
#
# Parameters:
# - name	a string: the name of the option to find
# - options	the list of options.
#
proc GetOption { name options } {

    set value {}

    # puts "Searching $name in $options"
    for { set i 0 } { $i < [llength $options] } { incr i 1 } {
	set opt [lindex $options $i]
	# puts "$name =? $opt"
	if ![string compare $name $opt] {
	    incr i 1
	    if { $i < [llength $options] } {
		set value [lindex $options $i]
		# puts "Found option $name: value $value"
	    }
	}
    }
    
    return $value
}

######################################################################
# TableMax - 
# A procedure to compute the maximal element of a range of a table.
# The table is assumed to be an array with numerical names and numerical
# entries.
#
# Parameters:
# - table	the table (an array) in question
# - ifrom	starting index
# - ito         ending index
#
proc TableMax { table ifrom ito } {
    upvar $table A
    # does not consider the possibility of an empty range.
    set res $A($ifrom)
    for { set i [expr $ifrom+1] } { $i <= $ito } { incr i 1 } {
	if { $A($i) > $res } { set res $A($i) }
    }

    return $res
}

######################################################################
# Remove_From_List -
# Function that removes elements containing a given tag at a given index
# from a list of lists. 
# For instance, if the list is l = [a 0 1] [b 0 2] [c 1 1], then
# "Remove_From_List 1 $l 0" will remove elements [a 0 1] and [b 0 2] 
# since they have a "0" in the position 1.
# 
# Parameters:
# - index	the index of each elements at which to search
# - the_list	the original list of elements
# - tag		the thing to search
#
proc Remove_From_List { index the_list tag } {

    set new_list { }
    foreach element $the_list {
	if { $tag != [lindex $element $index] } {
	    lappend new_list $element
	}
    }
    return $new_list
}

######################################################################
# Trim_Chordlist -
# Procedure to remove from a list of chords the ones that will create
# double arcs. Exception: if size/2 is a chord, it is left. 
# Also removes silly chord lengths: < 0 or > size.
#
# Parameters:
# - size	the total size of the "circle"
# - chordlist	a list of chords
#
proc Trim_Chordlist { size chordlist } {

    set newlist ""

    foreach val $chordlist {
	if [expr { $val >= 0 } && { $val <= $size }] {
	    if [expr 2*$val <= $size] {
		lappend newlist $val
	    } else {
		set compl [expr $size-$val]
		if [expr -1 == [lsearch -exact $chordlist $compl]] {
		    lappend newlist $val
		}
	    }
	}
    }

    return $newlist
}

######################################################################
# Extract_Braces -
# A function that extracts what is enclosed in braces in a string and 
# returns the result as a string. The nesting of braces is taken into 
# account so that the result on the string "a b { c { d } e f }" is 
# "c { d } e f".
#
# Parameters:
# - s		the string to process
#
proc Extract_Braces { s } {

    # puts "charcuting $s"
    set begin [string first \{ $s]
    if { $begin == -1 } { return "" }

    set ss [string range $s [expr $begin+1] end]
    set ss [string trimleft $ss]
    set sl [string length $ss]

    set CONTINUE 1
    set level 1
    for { set i 0 } { $CONTINUE } { incr i } {
	# puts -nonewline "|[string index $ss $i]+$level"
	if { [string index $ss $i] == "\{" } {
	    incr level
	} elseif { [string index $ss $i] == "\}" } {
	    incr level -1
	    if { $level == 0 } { 
		set CONTINUE 0
	    }
	}
	if { [expr $i + 1 ] >= $sl } {
	    set CONTINUE 0
	}
    }

    if { $level != 0 } { 
	return ""
    } else {
	# puts "-> [string trimright [string range $ss 0 [expr $i-2]]]"
	return [string trimright [string range $ss 0 [expr $i-2]]]
    }
	    
}

######################################################################
# isNumber -
# Boolean function checking if a string is a number
#
# Parameters:
# - number	the string to analyze
#
proc isNumber {number} {
    return [regexp {^ *[0-9]+ *$} $number]
}

######################################################################
# isReal -
# Boolean function checking if a string is a real number
#
# Parameters:
# - number	the string to analyze
#
proc isReal {number} {
    return [regexp {^ *[0-9]+\.?[0-9]* *$} $number]
}
