#
# Qualitative analysis module for pned
#
# Module grouping the procedures for "qualitative" analysis
# - checking topological properties
# - liveness
# - fifoness
# - emptyness
# - maxplusness
# - bounds on marking
#
######################################################################
# Check_Property -
# General handling procedure for calling a boolean function and displaying
# a message according to the result.
#
# Parameters:
# - test_function	the boolean function to call
# - title	title of the popup window
# - description	a string describing the property, used in the message.
#
proc Check_Property { test_function title description } {

    set res [$test_function]
    if { $res == 1 } {
	set message "This is $description."
    } elseif { $res == "unknown" } {
	set message "Can't tell if it is $description."
    } elseif { $res == 0 } {
	set message "This is not $description."
    }

    if { $res == 0 } {
	switch [tk_dialog .graph.property "Bad Properties" "$message\nDo you wish to see the bad objects?" {} 0 {Yes} {No}] {
	    0 { Display_Bad_Objects .graph.canvas }
	    1 { return }

	}
    } else {
	ReportInfo $title $message
    }
}

#######################################################################
# Check_Event_Graph -
# Check_Free_Choice -
# Check_State_Machine -
# Check_FIFO -
# Check_Liveness -
# Check_Deterministic -
# Check_Max_Plus -
# Front-end procedures to the corresponding boolean procedures
#
# Parameters: none
#
proc Check_Event_Graph { } {

    Check_Property Is_Event_Graph "Event Graph Analysis" "an Event Graph" 

}

proc Check_Free_Choice { } {

    Check_Property Is_Free_Choice "Free Choice Analysis" "a Free Choice Net"

}

proc Check_State_Machine { } {

    Check_Property Is_State_Machine "State Machine Analysis" "a State Machine"

}
    
proc Check_FIFO { } {

    Check_Property Is_FIFO "FIFO Analysis" "a FIFO system"

}

proc Check_Liveness { } {

    Check_Property Is_Live "Liveness Analysis" "a live net"

}

proc Check_Deterministic { } {

    Check_Property Is_FIFO "Deterministic Analysis" "a deterministic system"

}

proc Check_Max_Plus { } {

    Check_Property Is_Max_Plus "MaxPlus Analysis" "a (max,+) net"

}

######################################################################
# Is_FIFO -
# Quasi-Boolean function returning true if the Petri Net is FIFO.
# A net is FIFO when all objects are FIFO (see Has_FIFO_Objects) 
# or else if non-FIFO objects are in a cycle with one token. 
# This last possiblity is not tested yet.
#
# Parameters: none
#
proc Is_FIFO { } {

    # as long as timings are constant, the graph is FIFO.
    if [Has_FIFO_Objects] {
	return 1
    } else {
	# would need to test for safeness... no time for this now.
	return "unknown"
    }
}

######################################################################
# Is_Live -
# Quasi-Boolean function returning true if the Petri Net is live.
# This can be determined if the net is an Event Graph or a Free Choice
# Net by using eg_live and fc_live, respectively.
#
# Parameters: none
#
proc Is_Live { } {
    global KEEP bad_obj_list

    set TMP_EGL_FILE "/tmp/pned[pid].egl"
    set TMP_PNL_FILE "/tmp/pned[pid].pnl"
    set TMP_RES_FILE "/tmp/pned[pid].res"
    set TMP_MAT_FILE "/tmp/pned[pid].mat"

    if [Is_Event_Graph] {
	write_EGL $TMP_EGL_FILE 1
	if [file exists $TMP_RES_FILE] { 
	    RemoveFile $TMP_RES_FILE
	}

	if [catch "exec live_eg -f $TMP_EGL_FILE -d -o $TMP_RES_FILE" \
		problem] {
	    LogResult "Analysis Error" "" "$problem"
	    return
	}

	if [file exists $TMP_RES_FILE] {
	    set res 0
	    # getting the bad objects 
	    set resfile [open $TMP_RES_FILE r]
	    set objs {}
	    while {[gets $resfile line] >= 0} {
		if [expr { -1 == [string first "_" $line] } \
			&& { -1 == [string first "\#" $line] } \
			&& { 0 != [string length $line] } ] {
		    lappend objs $line
		}
	    }
	    close $resfile 
	    set bad_obj_list $objs
	} else {
	    set res 1
	}
    } elseif { [Is_Free_Choice] == 1 } {
	write_PNL $TMP_PNL_FILE

	exec fc2totmatrix -f $TMP_PNL_FILE -o $TMP_MAT_FILE

	# test if generation is ok: file should not begin with "The net..."
	set resfile [open $TMP_MAT_FILE r]
	set line [gets $resfile]
	close $resfile
	if [expr { [string index $line 1] != "T"} ] {
	    # ok. continuing

	    exec live_fc -f $TMP_MAT_FILE -o $TMP_RES_FILE -number

	    if [file exists $TMP_RES_FILE] {

		set resfile [open $TMP_RES_FILE r]
		set line [gets $resfile]
		if { $line == 0 } {
		    set res 1
		} else {
		    set res 0 
		}
		close $resfile
	    } else {
		ReportError "Analysis Error" "Network not in standard form."
		set res unknown
	    }
	} else {
	    ReportError "Analysis Error" "No Matrix. Liveness test failed."
	    set res unknown
	}
    } else {
	set res unknown
    }

    if !$KEEP {
	RemoveFile $TMP_EGL_FILE
	RemoveFile $TMP_PNL_FILE
	RemoveFile $TMP_MAT_FILE
	RemoveFile $TMP_RES_FILE
    }

    return $res
}

proc Compute_Bounds { } {
    global KEEP

    set TMP_PNL_FILE "/tmp/pned[pid].pnl"
    set TMP_RES_FILE "/tmp/pned[pid].res"

    write_PNL $TMP_PNL_FILE

    if [catch "exec pn_bound -f $TMP_PNL_FILE -d -o $TMP_RES_FILE" problem] {
	LogResult "Analysis Error" "" "$problem"
    } else {
	set result [open $TMP_RES_FILE r]
	set values [read $result]

	LogResult "Marking Bounds" "$values"
    }

    if !$KEEP {
	RemoveFile $TMP_PNL_FILE
	RemoveFile $TMP_RES_FILE
    }
}

######################################################################
# Is_Event_Graph -
# A boolean function answering true if the current Petri Net is an 
# Event Graph. The definition of an Event Graph is:
#  - all places have exactly 1 output and 1 input
#  - no weights on arcs.
#
# Parameters: none
#
proc Is_Event_Graph { } {
    global platra_list trapla_list place_list link_weight bad_obj_list 
    global DEBUG

    if [Is_Empty_Graph] { 
	set bad_obj_list {}
	return 0
    }

    # recompute place in/outdegrees
    foreach place $place_list {
	set place_id [lindex $place 0]
	set indegree($place_id) 0
	set outdegree($place_id) 0
    }
    foreach link $platra_list {
	set place_id [lindex $link 1]
	incr outdegree($place_id) 1
	if $DEBUG {
	    puts -nonewline "OD $place_id $outdegree($place_id) "
	}
    }
    foreach link $trapla_list {
	set place_id [lindex $link 2]
	incr indegree($place_id) 1
    }
    set ok 1
    set bad_obj_list {}
    foreach place $place_list {
	set place_id [lindex $place 0]
	if [expr { $indegree($place_id) != 1 } \
		|| { $outdegree($place_id) != 1 } ] {
	    set ok 0
	    set bad_obj_list [lappend bad_obj_list $place_id]
	}
    }
    # check link weights
    foreach link [concat $trapla_list $platra_list] {
	if { $link_weight([lindex $link 0]) != 1 } {
	    lappend bad_obj_list [lindex $link 0]
	    set ok 0
	}
    }

    if $DEBUG {
	puts "EG check: bad objects =\[ $bad_obj_list \]"
    }

    return $ok
}

######################################################################
# Is_Free_Choice -
# A boolean function answering true if the current Petri Net is a
# Free Choice Net. The definition of a Free Choice Net is:
#  - if a place has several output transitions, these transitions should
#    have exactly this place as input.
#  - no weights on arcs
#
# Parameters: none
#
proc Is_Free_Choice { } {
    global platra_list trapla_list trans_list place_list link_weight
    global bad_obj_list
    global DEBUG

    if [Is_Empty_Graph] { return 0 }

    # recompute trans indegrees and place outdegrees
    foreach trans $trans_list {
	set trans_id [lindex $trans 0]
	set indegree($trans_id) 0
    }
    foreach place $place_list {
	set place_id [lindex $place 0]
	set outdegree($place_id) 0
    }
    foreach link $platra_list {
	set place_id [lindex $link 1]
	set trans_id [lindex $link 2]
	incr indegree($trans_id) 1
	incr outdegree($place_id) 1
    }
    set ok 1
    set bad_obj_list {}
    foreach place $place_list {
	set place_id [lindex $place 0]
	if { $outdegree($place_id) > 1 } {
	    # there is a conflict. Scan links again
	    foreach link $platra_list {
		if { $place_id == [lindex $link 1] } {
		    if { $indegree([lindex $link 2]) > 1 } {
			set ok 0
			set bad_obj_list [lappend $bad_obj_list $place_id]
		    }
		}
	    }
	}
    }
    # check link weights
    foreach link [concat $trapla_list $platra_list] {
	if { $link_weight([lindex $link 0]) != 1 } {
	    lappend bad_obj_list [lindex $link 0]
	    set ok 0
	}
    }
    if $DEBUG {
	puts "FC check: bad objects: \[ $bad_obj_list \]"
    }

    return $ok
}

######################################################################
# Is_State_Machine -
# A boolean function answering true if the current Petri Net is a
# State Machine. The definition of a State_Machine is:
#  - each transition has exactly one input and one output
#  - no weights on arcs
#
# Parameters: none
#
proc Is_State_Machine { } {
    global platra_list trapla_list trans_list source_list bad_obj_list
    global link_weight
    global DEBUG

    if [Is_Empty_Graph] { return 0 }

    # recompute trans in/outdegrees
    foreach trans [concat $trans_list $source_list] {
	set trans_id [lindex $trans 0]
	set indegree($trans_id) 0
	set outdegree($trans_id) 0
    }
    foreach link $trapla_list {
	set trans_id [lindex $link 1]
	incr outdegree($trans_id) 1
    }
    foreach link $platra_list {
	set trans_id [lindex $link 2]
	incr indegree($trans_id) 1
    }
    set ok 1
    set bad_obj_list {}
    foreach trans [concat $trans_list $source_list] {
	set trans_id [lindex $trans 0]
	if [expr { $indegree($trans_id) != 1 } \
		|| { $outdegree($trans_id) != 1 } ] {
	    set ok 0
	    lappend bad_obj_list $trans_id
	}
    }
    # check link weights
    foreach link [concat $trapla_list $platra_list] {
	if { $link_weight([lindex $link 0]) != 1 } {
	    lappend bad_obj_list [lindex $link 0]
	    set ok 0
	}
    }
    if $DEBUG {
	puts "SM check: bad objects: \[ $bad_obj_list \]"
    }

    return $ok
}

######################################################################
# Contains_Sources -
# A boolean function answering true if the current Petri Net contains
# sources. 
#
# Parameters: none
#
proc Contains_Sources { } {
    global source_list

    return [expr [llength $source_list] > 0 ]
}

######################################################################
# Is_Empty_Graph -
# A boolean function answering true if the current Petri Net has no
# objects at all.
#
# Parameters: none
#
proc Is_Empty_Graph { } {
    global source_list place_list trans_list

    return [expr [llength $source_list]+[llength $trans_list]+[llength $place_list] == 0 ]
}

######################################################################
# Has_Fifo_Objects -
# A boolean function answering true if the current Petri Net has all
# its objects FIFO. A transition is FIFO if it is recycled or if it
# has a deterministic firing time. A place is FIFO if it has a 
# deterministic holding time.
#
# Parameters: none
#
proc Has_FIFO_Objects { } {
    global TransTime PlaceTime trans_list place_list bad_obj_list
    global DEBUG

    set ok 1
    set bad_obj_list {}

    foreach trans $trans_list {
	set trans_id [lindex $trans 0]
	array set Law $TransTime($trans_id)
	if { $Law(name) != "constant" } {
	    # look for recycling
	    set pl_ins [Get_Trans_Inputs $trans_id]
	    set ok 0
	    foreach i_place $pl_ins {
		set tr_ins [Get_Place_Inputs $i_place]
		puts "..t of $trans_id = $tr_ins"
		if { -1 != [lsearch $tr_ins $trans_id] } {
		    puts "found $trans_id in preceds"
		    set ok 1
		}
	    }
	    if !$ok {
		puts "BAD $trans_id"
		lappend bad_obj_list $trans_id
	    }
	}
    }

    foreach place $place_list {
	set place_id [lindex $place 0]
	array set Law $PlaceTime($place_id)
	if { $Law(name) != "constant" } {
	    set ok 0
	    lappend bad_obj_list $trans_id
	}
    } 

    # recomputing ok from the length of the bad object list
    set ok [expr 0 == [llength $bad_obj_list]]

    if $DEBUG {
	puts "FIFO check: bad objects: \[ $bad_obj_list \]"
    }

    return $ok
}
######################################################################
# Is_Deterministic -
# A boolean function answering true if the current Petri Net is
# deterministic, that is, has all timing (and routing, when available)
# laws constant.
#
# Parameters: none
#
proc Is_Deterministic { } {
    global TransTime PlaceTime trans_list source_list place_list bad_obj_list
    global DEBUG

    set ok 1
    set bad_obj_list {}

    foreach trans [concat $trans_list $source_list] {
	set trans_id [lindex $trans 0]
	array set Law $TransTime($trans_id)
	if { $Law(name) != "constant" } {
	    set ok 0
	    lappend bad_obj_list $trans_id
	}
    }

    foreach place $place_list {
	set place_id [lindex $place 0]
	array set Law $PlaceTime($place_id)
	if { $Law(name) != "constant" } {
	    set ok 0
	    lappend bad_obj_list $place_id
	}
    } 

    if $DEBUG {
	puts "DET check: bad objects: \[ $bad_obj_list \]"
    }

    return $ok
}

######################################################################
# Is_Max_Plus -
# A boolean function answering true if the current Petri Net is
# (max,+), that is, is an Event Graph and Deterministic.
#
# Parameters: none
#
proc Is_Max_Plus { } {
    global bad_obj_list
    global DEBUG

    # checking determinicity and saving the bad objects 
    set ok [Is_Deterministic]
    set badobj $bad_obj_list

    set ok [expr [Is_Event_Graph] && $ok]
    set bad_obj_list [concat $bad_obj_list $badobj]

    if $DEBUG {
	puts "MAX check: bad objects: \[ $bad_obj_list \]"
    }

    return $ok
}

######################################################################
# Is_Type -
# A boolean function answering true if the object passed is of the type
# passed. This is recognized by checking the presence of the tag "type"
# among the objects'.
#
# Parameters:
# - obj		the object to test
# - type	the type to check
#
proc Is_Type { obj type } {

    set objtags [.graph.canvas gettags $obj]
    return [expr  -1 != [lsearch -exact $objtags $type]]
}

######################################################################
# Has_Zero_Time -
# A boolean function answering true if the object passed has a timing
# with constant zero value.
#
# Parameters:
# - obj		the object to test
# - type	the type to check
#
proc Has_Zero_Time { obj } {
    global PlaceTime TransTime

    if [Is_Type $obj PLACE] {
	array set Law $PlaceTime($obj)
    } elseif [Is_Type $obj TRANS] {
	array set Law $TransTime($obj)
    } else { 
	return 0
    }

    return [Is_Zero_Time Law]
}

