#
# Module for creation and manipulation of basic objects of the editor.
# 
######################################################################
# Make_Place -
# A procedure to create a place with specified attributes.
# Options recognized are
# -name		optional, default ""
# -lawparams	optional, default to a constant distribution with value 0
# -x		mandatory, x position
# -y		mandatory, y position
#
#
# Parameters:
# - win		the graph canvas where the place is to be created
# - options	the list of options
#
proc Make_Task { win options } {
    global curzoomvalue task_radius TaskTime task_list name
    global state
    global DEBUG

    if $DEBUG {
	puts "Making a task with options $options"
    }

    set lawparams [list name constant value 0.0]
    set thename ""
    set scale [expr $curzoomvalue/100.0]

    for { set i 0 } { $i < [llength $options] } { incr i 1 } {
	set opt [lindex $options $i]
	switch -- $opt {
	    -lawparams { 
		incr i 1
		set lawparams [lindex $options $i]
	    }
	    -name { 
		incr i 1
		set thename [lindex $options $i]
	    }
	    -x {
		incr i 1
		set x [lindex $options $i]
	    }
	    -y {
		incr i 1
		set y [lindex $options $i]
	    }
	}
    }

    if ![expr { [info exists x] && [info exists y] }] {
	error "Create Error" "Could not create task: missing argument -x or -y."
    } else {
	# seems ok. Let's create.
	set newid [$win create oval \
		       [expr $x-$task_radius*$scale] \
		       [expr $y-$task_radius*$scale] \
		       [expr $x+$task_radius*$scale] \
		       [expr $y+$task_radius*$scale] \
		       -fill green -tag TASK]

	lappend task_list [list $newid $x $y 0]
    
	set TaskTime($newid) $lawparams
	set name($newid)     $thename

	Display_Name $win $newid
	Display_Timing $win $newid
	
	set state($win) modified

	return $newid
    }
}

######################################################################
# Put_Task -
# Procedure to put a new place at the specified position.
# The attributes of the new object are the default attributes.
# Used in the "place mode".
#
# Parameters:
# - win		the canvas where to put the object
# - x		x position
# - y 		y position
#
proc Put_Task { win x y } {
    global default_props

    set new_id [Make_Task $win \
		    [list \
			 -x $x \
			 -y $y \
			 -lawparams $default_props(TaskTime)
			] ]

    if { $new_id == "" } {
	error "Create Error" "Could not create place at ($x,$y)"
    } else {
	return $new_id
    }
}

proc Start_Link { win x y } {
    global curlinex curliney curlink
    global orig_type curzoomvalue task_radius
    global orig_obj
    global DEBUG

    set curobj [$win find withtag current]
    if $DEBUG {
	puts "pointing over $curobj"
    }

    set scale [expr $curzoomvalue/100.0]

    if { $curobj != "" } {
	set from [$win coords current]
	set objtags  [$win gettags $curobj]

	if {  -1 != [lsearch -exact $objtags TASK] } {
	    set orig_type TASK
	    set curlinex [expr [lindex $from 0]+$task_radius*$scale]
	    set curliney [expr [lindex $from 1]+$task_radius*$scale]
	}
	if { $orig_type != "" } {
	    .graph configure -cursor cross
	    if [info exists curlink] { unset curlink }
	    set orig_obj $curobj
	    bind $win <1> { Continue_Link %W [%W canvasx %x] [%W canvasy %y] }
	}
    }
}

proc Continue_Link { win x y } {
    global curlinex curliney curlink 
    global orig_type link_type curzoomvalue task_radius link_width

    # decide if it is the end ot not.
    if [Stop_Link $win $x $y] {
	return 
    }

    set scale [expr $curzoomvalue/100.0]

    if ![info exists curlink] {
	if { $orig_type == "TASK" } {
	    set end [clip $x $y $curlinex $curliney \
			 [expr $task_radius*$scale]]
	} else {
	    # no messy clipping for the time being
	    set end [list $curlinex $curliney]
	}
	set curlink [$win create line \
			 [lindex $end 0] [lindex $end 1] $x $y \
			 -tag {LINK $link_type} \
			 -width $link_width($link_type) ]
    } else {
	set ptlist [$win coords $curlink ]
	set ptlist [concat $ptlist [list $x $y]]
	eval {$win coords $curlink} $ptlist
    }

    set curlinex $x
    set curliney $y
}

proc Abort_Link { win } {
    global curlink

    if [info exists curlink] {
	$win delete $curlink
	unset curlink
    }
    bind $win <1> { Start_Link %W [%W canvasx %X] [%W canvasy %Y] }
}

proc Stop_Link { win x y } {
    global curlinex curliney curlink orig_type link_type link_lag LinkTime
    global curzoomvalue task_radius link_width create_props
    global orig_obj prog_link_list dep_link_list
    global DEBUG
    
    set cur_obj [$win find withtag current]
    if $DEBUG {
	puts "Current object = $cur_obj"
    }

    set scale [expr $curzoomvalue/100.0]

    if { "" != $cur_obj } {
	set objtags  [$win gettags $cur_obj]
	if $DEBUG {
	    puts " -> Has tags $objtags"
	}
	set goal [$win coords current]

	switch $orig_type {
	    TASK {
		if { -1 != [lsearch -exact $objtags "TASK"] } {
		    set end [clip $curlinex $curliney \
				 [expr [lindex $goal 0]+$task_radius*$scale] \
				 [expr [lindex $goal 1]+$task_radius*$scale] \
				 [expr $task_radius*$scale] ]
		    # but perhaps clipping of origin?
		    if ![info exists curlink] {
			set start [clip $x $y $curlinex $curliney \
				       [expr $task_radius*$scale]]
			set curlinex [lindex $start 0]
			set curliney [lindex $start 1]
		    }
		} else { return 0 }
	    }
	}

	set taglist [list LINK $link_type]
	if ![info exists curlink] {
	    set ptlist [list $curlinex $curliney \
			    [lindex $end 0] [lindex $end 1] ]
	    set curlink [eval {$win create line} $ptlist \
			     -tags {$taglist} -arrow last \
			     -width $link_width($link_type)]
	} else {
	    set ptlist [$win coords $curlink ]
	    set ptlist [concat $ptlist $end]
	    eval {$win coords $curlink} $ptlist
	    eval {$win itemconfigure $curlink} -tags {$taglist} -arrow last
	}

	# update of link lists
	switch $link_type {
	    PROG {
		set link_lag($curlink) $create_props(link_lag)
		set prog_link_list [lappend prog_link_list \
					[list $curlink $orig_obj $cur_obj]]
	    }
	    DEP {
		set LinkTime($curlink) $create_props(LinkTime)
		set dep_link_list [lappend dep_link_list \
					[list $curlink $orig_obj $cur_obj]]
	    }
	}

	.graph configure -cursor ""
	bind $win <1> { Start_Link %W [%W canvasx %X] [%W canvasy %Y] }
	bind $win <2> { Abort_Link %W }
	set orig_type ""

	unset curlink

	return 1
    } else {
	return 0
    }
}

######################################################################
# Delete_Object -
# Removes an object from the canvas. Removes also the "attached" objects
# such as text, links.
# Handles the redirection of the attention to the "owner" of the object
# in case the object is a text.
#
# Parameters:
# - win		the canvas where this is happening
# - obj_id	the id of the object to remove with its "environment"
#
proc Delete_Object { win obj_id } {
    global task_list dep_link_list prog_link_list
    global DEBUG

    if { "" == $obj_id } { return }

    set objtags  [$win gettags $obj_id]
    
    if $DEBUG {
	puts "DEL: object = $obj_id, tags = $objtags"
    }

    if [Is_Type $obj_id TASK] {
	set task_list [Remove_From_List 0 $task_list $obj_id]
	Delete_From_List $win 1 $dep_link_list $obj_id
	Delete_From_List $win 2 $dep_link_list $obj_id
	Delete_From_List $win 1 $prog_link_list $obj_id
	Delete_From_List $win 2 $prog_link_list $obj_id
	set dep_link_list [Remove_From_List 1 $dep_link_list $obj_id]
	set dep_link_list [Remove_From_List 2 $dep_link_list $obj_id]
	set prog_link_list [Remove_From_List 1 $prog_link_list $obj_id]
	set prog_link_list [Remove_From_List 2 $prog_link_list $obj_id]
    } elseif [Is_Type $obj_id LINK] {
	if [Is_Type $obj_id DEP] {
	    set dep_link_list [Remove_From_List 0 $dep_link_list $obj_id]
	} elseif [Is_Type $obj_id PROG] {
	    set prog_link_list [Remove_From_List 0 $prog_link_list $obj_id]
	}
    } elseif [Is_Type $obj_id TEXT] {
	# might be text. Can't delete it directly. Have to find to whom it
	# belongs.
	foreach tag $objtags {
	    if [expr [string match "Tx*" $tag] \
		    || [string match "Ti*" $tag] ] {
		set new_obj_id [string range $tag 2 end]
	    }
	}
	if [info exists new_obj_id] {
	    Delete_Object $win $new_obj_id
	}
	# whatever the result, return.
	return
    }

    # if selection, should be destroyed with its object
    if ![Is_Type $obj_id SELECT] {
	$win delete $obj_id

	# delete text associated
	$win delete [.graph.canvas find withtag "Tx$obj_id"]

	# delete text timing associated
	$win delete [.graph.canvas find withtag "Ti$obj_id"]

	# delete "selection" if any
	$win delete [.graph.canvas find withtag "Sel$obj_id"]
    }
}

#######################################################################
# Delete_Current_Object -
# Removes the "current" object, usually the one pointed at. 
#
# Parameters:
# - win		the canvas where this is happening
#
proc Delete_Current_Object { win } {
    global DEBUG

    set cur_obj [$win find withtag current]

    if $DEBUG {
	puts "Current object = $cur_obj"
    }

    if { [llength $cur_obj] > 1 } { 
	puts "Warning! More than one current object!"
    }
    if { [llength $cur_obj] > 0 } { 
	Delete_Object $win $cur_obj
    }
}

